// this is a new start on a suballocator, this will be a combination of
// Borland's suballocator for all versions before version 4.02 and some extra
// code so I can keep track of free spaces in a particular block.
// an allocation has a two byte header that indicates the length of the
// allocation, a free space has the length followed by previous free and next
// free. As far as I could tell, the free list was in no particular order in
// Borland's suballocator, mine will be in sorted order after it is
// compressed.
#ifdef SUBALLOC
#include <dos.h>
#include <mem.h>
#include <alloc.h>

static char **Segments=NULL;
static short int Segment_Count;
static short int Compact_Flag=0;
typedef struct
{
	short int Segment;
	unsigned short int Size;
} byte_search;

static byte_search Found_Block;
static byte_search Found_Empty_Block;

static short near Alloc_Segment(void)
{
	char *segment;
	if(Segment_Count == 0)
		Segments = (char **)malloc(sizeof(char *)*257);
	if(segment = Segments[Segment_Count] = (char *)calloc(1,65535),
		segment == NULL)
		return 0;

	// offset to first free in this segment.
	*((unsigned short *)segment+1) = 0;
	Segment_Count++;
	return 1;
}

static short int near Expand_Block(void *ptr, unsigned short int new_len)
{
	short int i;
	unsigned short int seg2;
	unsigned short int seg = FP_SEG(ptr);
	unsigned short int next_alloc;
	unsigned short int max_expand;
	unsigned short int len;

	// the first source for allocations is freed space.
	for(i = 0;i<Segment_Count;i++)
	{
		seg2 = FP_SEG(Segments[i]);
		if(seg == seg2)
		{
			unsigned short tmp;
			len = _msize(ptr);
			tmp = FP_OFF(ptr) + len;

			// position 0 in each segment allocation tells where the end of
			// allocations is.
			next_alloc = *(unsigned short int *)MK_FP(seg2,0);

			// if this segment is not completely full and this allocation is at
			// the end of the list of allocations.
			if(next_alloc && tmp == next_alloc)
			{
				max_expand = 65535 - next_alloc;
				// if the expansion can take place without moving data...
				if(len + max_expand > new_len)
				{
					*((unsigned short int *)ptr - 1) = new_len;
					*(unsigned short int *)MK_FP(seg2,0) =
						FP_OFF(ptr) + new_len;
					return 1;
				}
			}
			else
			{
				unsigned short int noffset;
				unsigned short int next_free = *(unsigned short int *)
					MK_FP(seg2,sizeof(unsigned short int));
				while(next_free)
				{
					noffset = *(unsigned short int *)
						MK_FP(seg2,next_free+2*sizeof(unsigned short int));
					// see if the next block is in the free list for this segment
					if(tmp == next_free)
					{
						unsigned short int poffset;
						unsigned short int flen;
						poffset = *(unsigned short int *)
							MK_FP(seg2,next_free+sizeof(unsigned short int));

						flen = *(unsigned short int *)MK_FP(seg2,next_free);

						if(flen + len + sizeof(unsigned short int) >= new_len)
						{
							if(flen + len + sizeof(unsigned short int) < new_len + 8)
							{
								// remove this free space from the list.
								if(poffset)
									*(unsigned short int *)MK_FP(seg2,poffset+2*sizeof(unsigned short int)) = noffset;
								else
									*(unsigned short int *)MK_FP(seg2,sizeof(unsigned short int)) = noffset;

								if(noffset)
									*(unsigned short int *)MK_FP(seg2,noffset+sizeof(unsigned short int)) = poffset;
								*((unsigned short int *)ptr - 1) = flen + len + sizeof(unsigned short int);
							}
							else
							{
								// create a new, smaller free space at the end of the
								// space the block is expanding into.
								unsigned short int new_free = next_free + new_len - len;
								unsigned short int *fr =
									(unsigned short int *)MK_FP(seg2,new_free);
								*fr = flen - (new_len - len);
								fr++;
								*fr = poffset;
								fr++;
								*fr = noffset;
								*((unsigned short int *)ptr - 1) = new_len;

								if(poffset)
									*(unsigned short int *)MK_FP(seg2,poffset+2*sizeof(unsigned short int)) = new_free;
								else
									*(unsigned short int *)MK_FP(seg2,sizeof(unsigned short int)) = new_free;

								if(noffset)
									*(unsigned short int *)MK_FP(seg2,noffset+sizeof(unsigned short int)) = new_free;
							}
							return 1;
						}
						return 0;
					}

					next_free = noffset;
				}
			}
			break;
		}
	}
	return 0;
}

typedef struct
{
	unsigned short int Offset;
	unsigned short int Len;
	int Change;
} Free_List;

static int list_sort(const void *p1, const void *p2)
{
	long ret;

	ret = ((Free_List *)p1)->Offset - ((Free_List *)p2)->Offset;
	return (short int)ret;
}

#define COMPACT_FREE_LIST
static void near Compact_Free_List(void)
{
	short int i,j;
	unsigned short int seg;
	short int changed=0;
	short int ch = 0;
	short int free_count=0;
	short int max=0;
	Free_List *list=NULL;
	unsigned short int next_free;

	// combine any free spaces that are contiguous.
	for(i = 0;i<Segment_Count;i++)
	{
		seg = FP_SEG(Segments[i]);
		// get first free.
		next_free = *(unsigned short int *)MK_FP(seg,sizeof(unsigned short int));

		// make a list of all free spaces
		if(next_free)
		{
			Free_List *cur;

			list = (Free_List *)realloc(list,sizeof(Free_List)*10922);

			do
			{
				// add this free space to the list.
				list[free_count].Offset = next_free;
				list[free_count].Len = *(unsigned short int *)MK_FP(seg,next_free);
				free_count++;

				// get next free.
				next_free = *(unsigned short int *)
					MK_FP(seg,next_free+2*sizeof(unsigned short int));
			} while(next_free);

			// sort by offset.
			qsort(list,free_count,sizeof(Free_List),list_sort);

			// combine any that are contiguous.
			for(j = 0;j<free_count-1;j++)
			{
				Free_List *nxt = &list[j+1];
				cur = &list[j];
				// if this and the next are contiguous, combine and remove the next one.
				if(cur->Offset+cur->Len+sizeof(unsigned short int) == nxt->Offset)
				{
					// if any were combined, set changed flag.
					changed = 1;
					cur->Change = 1;
					cur->Len += nxt->Len + sizeof(unsigned short int);
					memmove(nxt,nxt+1,sizeof(Free_List)*(free_count - j - 2));  // is this length right?
					free_count--;
					j--;
				}
			}

			if(changed)
			{
				unsigned short int *fr;
				// write out new offsets in sorted order.
				// write first offset into segment header.
				*(unsigned short int *)MK_FP(seg,sizeof(unsigned short int)) = list->Offset;
				fr = (unsigned short int *)MK_FP(seg,list->Offset);
				if(list->Change)
					*fr = list->Len;
				fr++;
				// its previous is 0
				*fr = 0;
				fr++;
				if(free_count == 1)
				{
					// this case is if this is now the only free space.
					*fr = 0;
					goto end;
				}
				else
					*fr = list[1].Offset;

				// all except the last are easy.
				for(j = 1;j<free_count-1;j++)
				{
					cur = &list[j];
					fr = (unsigned short int *)MK_FP(seg,cur->Offset);
					if(cur->Change)
						*fr = cur->Len;
					fr++;
					*fr = list[j-1].Offset;
					fr++;
					*fr = list[j+1].Offset;
				}

				// last one's previous is obvious, its next is 0.
				fr = (unsigned short int *)MK_FP(seg,list[free_count-1].Offset);
				if(list[free_count-1].Change)
					*fr = list[free_count-1].Len;
				fr++;
				*fr = list[free_count-2].Offset;
				fr++;
				*fr = 0;
				ch = 1;
			}

			end:
			// a later addition would be to combine the last free space with the
			// unallocated space at the end of a segment if they are contiguous.
			// that would be done right here. If this is done, it would also be
			// possible to end up with a segment that is entirely free space and
			// could be removed from the list.
			if(free_count)
			{
				Free_List *prv = &list[free_count-1];
				// using next free because it's here, next_alloc would be the
				// correct term.
				next_free = *(unsigned short int *)MK_FP(seg,0);
				// if the end of the allocated space is right after the last free
				if(prv->Offset + prv->Len == next_free)
				{
					// combine the two.
					unsigned short int *prev,*fr;
					*(unsigned short int *)MK_FP(seg,0) = prv->Offset;
					fr = (unsigned short int *)MK_FP(seg,prv->Offset);
					fr++;
					if(*fr)
					{
						prev = (unsigned short int *)MK_FP(seg,*fr);
						prev++; // skip over length.
						prev++; // skip over previous
						*prev = 0; // change next to empty.
					}
					else
						*(unsigned short int *)MK_FP(seg,sizeof(unsigned short)) = 0;
				}
			}

			if(list)
			{
				free(list);
				list = NULL;
			}
			max = 0;
			free_count = 0;
			changed = 0;
		}
	}

	if(ch)
	{
		Found_Block.Size = 0;
		Found_Block.Segment = 0;
		Found_Empty_Block.Size = 0;
		Found_Empty_Block.Segment = 0;
	}
}

static short int count = 0;

static void * Find_Free(size_t size)
{
	unsigned short int seg;
	unsigned short int len;
	unsigned short int noffset;
	short int i;
	unsigned short int free_offset;
	unsigned short int next_alloc;
	#ifdef COMPACT_FREE_LIST
	void *ptr;
	#endif

	count++;
	if(count >= 20000)
	{
		Compact_Free_List();
		count = 0;
	}

	if(size < 6)
		size = 6;

	i = 0;
	// this code will eliminate the need for the free search to start clear
	// at the start of the list on a frequent search for a given size.
	if(Found_Block.Size <= size)
		i = Found_Block.Segment;

	// the first source for allocations is freed space.
	for(;i<Segment_Count;i++)
	{
		seg = FP_SEG(Segments[i]);
		free_offset = *(unsigned short *)MK_FP(seg,sizeof(unsigned short int));
		while(free_offset)
		{
			len = *(unsigned short int *)MK_FP(seg,free_offset);
			noffset = *(unsigned short int *)MK_FP(seg,free_offset+2*sizeof(unsigned short int));
			if(len >= size)
			{
				unsigned short int poffset;
				poffset = *(unsigned short int *)MK_FP(seg,free_offset+sizeof(unsigned short int));

				if(len < size + 8)
				{
					if(poffset)
						*(unsigned short int *)MK_FP(seg,poffset+2*sizeof(unsigned short int)) = noffset;
					else
						*(unsigned short int *)MK_FP(seg,sizeof(unsigned short int)) = noffset;

					if(noffset)
						*(unsigned short int *)MK_FP(seg,noffset+sizeof(unsigned short int)) = poffset;
				}
				else
				{
					unsigned short int new_free = free_offset + size + sizeof(unsigned short int);
					unsigned short int *fr = (unsigned short int *)MK_FP(seg,new_free);
					*fr = len - size - sizeof(unsigned short int);
					fr++;
					*fr = poffset;
					fr++;
					*fr = noffset;
					fr = (unsigned short int *)MK_FP(seg,free_offset);
					*fr = size;

					if(poffset)
						*(unsigned short int *)MK_FP(seg,poffset+2*sizeof(unsigned short int)) = new_free;
					else
						*(unsigned short int *)MK_FP(seg,sizeof(unsigned short int)) = new_free;

					if(noffset)
						*(unsigned short int *)MK_FP(seg,noffset+sizeof(unsigned short int)) = new_free;
				}
				Found_Block.Size = size;
				Found_Block.Segment = i;
				return MK_FP(seg,free_offset+sizeof(unsigned short int));
			}
			free_offset = noffset;
		}
	}
	Found_Block.Size = size;
	Found_Block.Segment = Segment_Count;

	// the second source of allocation material is ends of segments
	i = 0;
	if(size >= Found_Empty_Block.Size)
		i = Found_Empty_Block.Segment;

	for(;i<Segment_Count;i++)
	{
		seg = FP_SEG(Segments[i]);
		next_alloc = *(unsigned short *)MK_FP(seg,0);
		len = 65535 - next_alloc;
		if(len >= size + sizeof(unsigned short int))
		{
			Found_Empty_Block.Size = size;
			Found_Empty_Block.Segment = i;
			*(unsigned short int *)MK_FP(seg,next_alloc) = size;
			*(unsigned short *)MK_FP(seg,0) = next_alloc + sizeof(unsigned short) + size;
			return MK_FP(seg,next_alloc+sizeof(unsigned short int));
		}
	}

	// the last source for allocation material is to create a new segment.
	if(Alloc_Segment())
	{
		short int segment = Segment_Count-1;
		seg = FP_SEG(Segments[segment]);
		next_alloc = 4;
		Found_Empty_Block.Size = size;
		Found_Empty_Block.Segment = segment;
		*(unsigned short int *)MK_FP(seg,next_alloc) = size;
		*(unsigned short *)MK_FP(seg,0) = next_alloc + sizeof(unsigned short) + size;
		return MK_FP(seg,next_alloc+sizeof(unsigned short int));
	}

	// if these three methods failed, we are really and truly out of options.
	// try compacting the free list and trying it again.
	#ifdef COMPACT_FREE_LIST
	ptr = NULL;
	if(!Compact_Flag)
	{
		Compact_Flag = 1;
		Compact_Free_List();
		ptr = Find_Free(size);
		Compact_Flag = 0;
	}
	return ptr;
	#else
	return NULL;
	#endif
}

void * _loadds _export Malloc(size_t size)
{
	return Find_Free(size);
}

void _loadds _export Free(void *block)
{
	short int i;
	unsigned short int seg;
	unsigned short int test = FP_SEG(block);
	unsigned short int first_free;
	char *ptr;

	for(i = 0;i<Segment_Count;i++)
	{
		seg = FP_SEG(Segments[i]);
		if(seg == test)
		{
			short int len = _msize(block);
			unsigned short int offset;
			if(len >= Found_Block.Size && i < Found_Block.Segment)
				Found_Block.Segment = i;

			first_free = *(unsigned short int *)MK_FP(seg,sizeof(unsigned short int));
			offset = FP_OFF(block) - sizeof(unsigned short int);

			if(first_free)
			{
				// set previous in what used to be first free to this block.
				*(unsigned short int *)
					MK_FP(seg,first_free+sizeof(unsigned short int)) =
					offset;

				*((unsigned short int *)block+1) = first_free; // former first is now next
			}
			else
				*((unsigned short int *)block+1) = 0;
			*(unsigned short int *)block = 0; // no previous
			*(unsigned short int *)MK_FP(seg,sizeof(unsigned short int)) = offset;
			break;
		}
	}
}

void * _loadds _export Realloc(void *ptr, size_t size)
{
	unsigned short int len;
	char *tmp = ptr;
	if(!ptr)
		return Malloc(size);

	len = _msize(ptr);
	if(len >= size)
		return ptr;
	else
	{
		void *nw;
		if(!Expand_Block(ptr,size))
		{
			nw = Malloc(size);
			memcpy(nw,ptr,len);
			Free(ptr);
			return nw;
		}
		return ptr;
	}
}

char * _loadds _export Strdup(char *string)
{
	short int len = strlen(string);
	char *ptr = (char *)Malloc(len+1);
	memcpy(ptr,string,len+1);
	return ptr;
}

void * _loadds _export Calloc(size_t size, size_t count)
{
	void *ptr = Find_Free(size*count);
	memset(ptr,0,size*count);
	return ptr;
}

unsigned short int _msize(void *ptr)
{
	return *((unsigned short int *)ptr-1);
}

#ifdef __cplusplus
// it is necessary to either rename this file to a ".CPP" extension or to
// force a C++ compile on this file to get this code included.

// this is an overloaded new operator that will replace the normal one
void * operator new(size_t size)
{
	// allocate the room
	return Malloc(size);
}

// this is an overloaded delete operator that replaces the normal one.
void operator delete(void *ptr)
{
	// if they pass in a pointer
	if(ptr)
		Free(ptr);
}
#endif
#endif
